@@ -1,6 +1,6 @@
 function _instanceof(left, right) {
   if (right != null && typeof Symbol !== "undefined" && right[Symbol.hasInstance]) {
-    return right[Symbol.hasInstance](left);
+    return !!right[Symbol.hasInstance](left);
   } else {
     return left instanceof right;
   }
